'''
Main modules BacterialTyper
'''
__all__ = [
	'annot',
	'assemble',
	'citation',
	'cluster',
	'config',
	'database',
	'ident',
	'help_info',
	'MGE',
	'phylo',
	'profile',
	'prep',
	'qc',
	'run',
	'report_generation',
	'test',
	'trimm',
	'version'
]

from BacterialTyper.modules import *
